module ListTests exposing (all)

import Expect
import Fuzz
import List.Extra exposing (Step(..))
import Test exposing (Test, describe, fuzz, fuzz2, fuzz3, test)


all : Test
all =
    describe "List.Extra"
        [ describe "unique" <|
            [ test "removes duplicates" <|
                \() ->
                    Expect.equal (List.Extra.unique [ 0, 1, 1, 0, 1 ]) [ 0, 1 ]
            , test "preserves list order" <|
                \() ->
                    Expect.equal (List.Extra.unique [ 3, 3, 2, 1, 1, 0 ]) [ 3, 2, 1, 0 ]
            ]
        , describe "allDifferent" <|
            [ test "detects duplicates" <|
                \() ->
                    Expect.equal (List.Extra.allDifferent [ 0, 1, 1, 0, 1 ]) False
            ]
        , describe "andMap" <|
            [ test "computes piecemeal" <|
                \() ->
                    Expect.equal
                        ([ 1, 2, 3 ]
                            |> List.map (\a b c -> a + b * c)
                            |> List.Extra.andMap [ 4, 5, 6 ]
                            |> List.Extra.andMap [ 2, 1, 1 ]
                        )
                        [ 9, 7, 9 ]
            ]
        , describe "reverseMap" <|
            [ test "maps and reverses" <|
                \() ->
                    Expect.equal (List.Extra.reverseMap sqrt [ 1, 4, 9 ]) [ 3, 2, 1 ]
            ]
        , describe "notMember" <|
            [ test "disconfirms member" <|
                \() ->
                    Expect.equal (List.Extra.notMember 1 [ 1, 2, 3 ]) False
            , test "confirms non-member" <|
                \() ->
                    Expect.equal (List.Extra.notMember 4 [ 1, 2, 3 ]) True
            ]
        , describe "find" <|
            [ test "behaves as documented" <|
                \() ->
                    Expect.equal (List.Extra.find (\num -> num > 5) [ 2, 4, 6, 8 ]) (Just 6)
            ]
        , describe "elemIndex" <|
            [ test "finds index of value" <|
                \() ->
                    Expect.equal (List.Extra.elemIndex 1 [ 1, 2, 3 ]) (Just 0)
            , test "doesn't find index of non-present" <|
                \() ->
                    Expect.equal (List.Extra.elemIndex 4 [ 1, 2, 3 ]) Nothing
            , test "finds index of first match" <|
                \() ->
                    Expect.equal (List.Extra.elemIndex 1 [ 1, 2, 1 ]) (Just 0)
            ]
        , describe "elemIndices" <|
            [ test "finds singleton index" <|
                \() ->
                    Expect.equal (List.Extra.elemIndices 1 [ 1, 2, 3 ]) [ 0 ]
            , test "doesn't find indices of non-present" <|
                \() ->
                    Expect.equal (List.Extra.elemIndices 4 [ 1, 2, 3 ]) []
            , test "finds all indices" <|
                \() ->
                    Expect.equal (List.Extra.elemIndices 1 [ 1, 2, 1 ]) [ 0, 2 ]
            ]
        , describe "findIndex" <|
            [ test "finds index of value" <|
                \() ->
                    Expect.equal (List.Extra.findIndex (\x -> modBy 2 x == 0) [ 1, 2, 3 ]) (Just 1)
            , test "doesn't find index of non-present" <|
                \() ->
                    Expect.equal (List.Extra.findIndex (\x -> modBy 2 x == 0) [ 1, 3, 5 ]) Nothing
            , test "finds index of first match" <|
                \() ->
                    Expect.equal (List.Extra.findIndex (\x -> modBy 2 x == 0) [ 1, 2, 4 ]) (Just 1)
            ]
        , describe "findIndices" <|
            [ test "finds singleton index" <|
                \() ->
                    Expect.equal (List.Extra.findIndices (\x -> modBy 2 x == 0) [ 1, 2, 3 ]) [ 1 ]
            , test "doesn't find indices of non-present" <|
                \() ->
                    Expect.equal (List.Extra.findIndices (\x -> modBy 2 x == 0) [ 1, 3, 5 ]) []
            , test "finds all indices" <|
                \() ->
                    Expect.equal (List.Extra.findIndices (\x -> modBy 2 x == 0) [ 1, 2, 4 ]) [ 1, 2 ]
            ]
        , describe "findMap" <|
            [ test "Nothing for empty list" <|
                \() ->
                    Expect.equal (List.Extra.findMap identity []) Nothing
            , test "Finds and maps for list of one with first element matching" <|
                \() ->
                    Expect.equal (List.Extra.findMap List.head [ [ 1 ] ]) (Just 1)
            , test "Fails to find and map for list of one without element matching" <|
                \() ->
                    Expect.equal (List.Extra.findMap List.head [ [] ]) Nothing
            , test "Finds and maps for list with middle element matching" <|
                \() ->
                    Expect.equal (List.Extra.findMap List.head [ [], [ 2 ], [] ]) (Just 2)
            , test "Finds and maps for list with last element matching" <|
                \() ->
                    Expect.equal (List.Extra.findMap List.head [ [], [], [ 3 ] ]) (Just 3)
            , test "Fails to find and map for list with no element matching" <|
                \() ->
                    Expect.equal (List.Extra.findMap List.head [ [], [], [] ]) Nothing
            ]
        , describe "count" <|
            [ test "isOdd predicate" <|
                \() ->
                    Expect.equal (List.Extra.count (\n -> modBy 2 n == 1) [ 1, 2, 3, 4, 5, 6, 7 ]) 4
            , test "equal predicate" <|
                \() ->
                    Expect.equal (List.Extra.count ((==) "yeah") [ "She", "loves", "you", "yeah", "yeah", "yeah" ]) 3
            ]
        , describe "intercalate" <|
            [ test "computes example" <|
                \() ->
                    Expect.equal
                        (List.Extra.intercalate [ 0, 0 ] [ [ 1, 2 ], [ 3, 4 ], [ 5, 6 ] ])
                        [ 1, 2, 0, 0, 3, 4, 0, 0, 5, 6 ]
            ]
        , describe "transpose" <|
            [ test "performs basic transpose" <|
                \() ->
                    Expect.equal
                        (List.Extra.transpose [ [ 1, 2, 3 ], [ 4, 5, 6 ] ])
                        [ [ 1, 4 ], [ 2, 5 ], [ 3, 6 ] ]
            , test "truncate the matrix to the shortest row size" <|
                \() ->
                    Expect.equal
                        (List.Extra.transpose [ [ 10, 11 ], [ 20 ], [ 30, 31, 32 ] ])
                        [ [ 10, 20, 30 ] ]
            , test "transposes large lists" <|
                \() ->
                    Expect.equal
                        (List.Extra.transpose [ List.repeat 10000 1 ])
                        (List.repeat 10000 [ 1 ])
            ]
        , describe "subsequences" <|
            [ test "computes subsequences" <|
                \() ->
                    Expect.equal
                        (List.Extra.subsequences [ 1, 2, 3 ])
                        [ [], [ 1 ], [ 2 ], [ 1, 2 ], [ 3 ], [ 1, 3 ], [ 2, 3 ], [ 1, 2, 3 ] ]
            ]
        , describe "subsequencesNonEmpty" <|
            [ test "computes subsequences to non-empty lists" <|
                \() ->
                    Expect.equal
                        (List.Extra.subsequencesNonEmpty [ 1, 2, 3 ])
                        [ ( 1, [] )
                        , ( 2, [] )
                        , ( 1, [ 2 ] )
                        , ( 3, [] )
                        , ( 1, [ 3 ] )
                        , ( 2, [ 3 ] )
                        , ( 1, [ 2, 3 ] )
                        ]
            ]
        , describe "permutations" <|
            [ test "computes permutations" <|
                \() ->
                    Expect.equal
                        (List.Extra.permutations [ 1, 2, 3 ])
                        [ [ 1, 2, 3 ], [ 1, 3, 2 ], [ 2, 1, 3 ], [ 2, 3, 1 ], [ 3, 1, 2 ], [ 3, 2, 1 ] ]
            ]
        , describe "isPermutationOf"
            [ fuzz2 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "works the same as sorting" <|
                \a b ->
                    List.Extra.isPermutationOf a b
                        |> Expect.equal (List.sort a == List.sort b)
            , test "correctly notices permutations" <|
                \() ->
                    Expect.all
                        ([ [ 1, 2, 3 ], [ 1, 3, 2 ], [ 2, 1, 3 ], [ 2, 3, 1 ], [ 3, 1, 2 ], [ 3, 2, 1 ] ]
                            |> List.map
                                (\permutation () ->
                                    permutation
                                        |> List.Extra.isPermutationOf [ 1, 2, 3 ]
                                        |> Expect.equal True
                                )
                        )
                        ()
            , test "Notices that 1,1,1 is not a permutation of 1,2,3" <|
                \() ->
                    List.Extra.isPermutationOf [ 1, 1, 1 ] [ 1, 2, 3 ]
                        |> Expect.equal False
            , test "Notices that 1,1,2 is not a permutation of 1,2,2" <|
                \() ->
                    List.Extra.isPermutationOf [ 1, 1, 2 ] [ 1, 2, 2 ]
                        |> Expect.equal False
            , test "correctly notices non-permutations" <|
                \() ->
                    Expect.all
                        ([ [], [ 1, 3 ], [ 2, 1, 3, 2 ], [ 4, 3, 1 ] ]
                            |> List.map
                                (\nonPermutation () ->
                                    [ 1, 2, 3 ]
                                        |> List.Extra.isPermutationOf nonPermutation
                                        |> Expect.equal False
                                )
                        )
                        ()
            ]
        , describe "interweave" <|
            [ test "interweaves lists of equal length" <|
                \() ->
                    Expect.equal (List.Extra.interweave [ 1, 3 ] [ 2, 4 ]) [ 1, 2, 3, 4 ]
            , test "appends remaining members of first list, if longer" <|
                \() ->
                    Expect.equal (List.Extra.interweave [ 1, 3, 5, 7 ] [ 2, 4 ]) [ 1, 2, 3, 4, 5, 7 ]
            , test "appends remaining members of second list, if longer" <|
                \() ->
                    Expect.equal (List.Extra.interweave [ 4, 9, 16 ] [ 2, 3, 5, 7 ]) [ 4, 2, 9, 3, 16, 5, 7 ]
            ]
        , describe "cartesianProduct" <|
            [ test "computes the cartesian product of lists of different length" <|
                \() ->
                    Expect.equal (List.Extra.cartesianProduct [ [ 1, 2 ], [ 3, 4, 5 ], [ 6 ] ]) [ [ 1, 3, 6 ], [ 1, 4, 6 ], [ 1, 5, 6 ], [ 2, 3, 6 ], [ 2, 4, 6 ], [ 2, 5, 6 ] ]
            , test "computes the cartesian product of a single list" <|
                \() ->
                    Expect.equal (List.Extra.cartesianProduct [ [ 1, 2 ] ]) [ [ 1 ], [ 2 ] ]
            , test "computes the cartesian product of lists including an empty one" <|
                \() ->
                    Expect.equal (List.Extra.cartesianProduct [ [ 1, 2 ], [], [ 6 ] ]) []
            , test "computes the nullary cartesian product" <|
                \() ->
                    Expect.equal (List.Extra.cartesianProduct []) [ [] ]
            ]
        , describe "foldl1" <|
            [ test "computes minimum" <|
                \() ->
                    Expect.equal (List.Extra.foldl1 min [ 1, 2, 3, 2, 1 ]) (Just 1)
            , test "computes left to right difference" <|
                \() ->
                    Expect.equal (List.Extra.foldl1 (-) [ 1, 2, 3, 4 ]) (Just 2)
            , test "concats in reverse" <|
                \() ->
                    Expect.equal (List.Extra.foldl1 (++) [ "a", "b", "c" ]) (Just "cba")
            , test "falls back to Nothing" <|
                \() ->
                    Expect.equal (List.Extra.foldl1 min []) Nothing
            ]
        , describe "foldr1" <|
            [ test "computes minimum" <|
                \() ->
                    Expect.equal (List.Extra.foldr1 min [ 1, 2, 3, 2, 1 ]) (Just 1)
            , test "computes right to left difference" <|
                \() ->
                    Expect.equal (List.Extra.foldr1 (-) [ 1, 2, 3, 4 ]) (Just -2)
            , test "concats properly" <|
                \() ->
                    Expect.equal (List.Extra.foldr1 (++) [ "a", "b", "c" ]) (Just "abc")
            , test "falls back to Nothing" <|
                \() ->
                    Expect.equal (List.Extra.foldr1 min []) Nothing
            ]
        , describe "scanl1" <|
            [ test "computes left to right iterative sum" <|
                \() ->
                    Expect.equal (List.Extra.scanl1 (+) [ 1, 2, 3 ]) [ 1, 3, 6 ]
            , test "computes left to right iterative difference" <|
                \() ->
                    Expect.equal (List.Extra.scanl1 (-) [ 1, 2, 3 ]) [ 1, 1, 2 ]
            , test "computes left to right flipped iterative difference" <|
                \() ->
                    Expect.equal (List.Extra.scanl1 (\x y -> y - x) [ 1, 2, 3 ]) [ 1, -1, -4 ]
            ]
        , describe "scanr" <|
            [ test "computes right to left iterative sum" <|
                \() ->
                    Expect.equal (List.Extra.scanr (+) 0 [ 1, 2, 3 ]) [ 6, 5, 3, 0 ]
            , test "computes right to left iterative difference" <|
                \() ->
                    Expect.equal (List.Extra.scanr (-) 0 [ 1, 2, 3 ]) [ 2, -1, 3, 0 ]
            ]
        , describe "scanr1" <|
            [ test "computes right to left iterative sum" <|
                \() ->
                    Expect.equal (List.Extra.scanr1 (+) [ 1, 2, 3 ]) [ 6, 5, 3 ]
            , test "computes right to left iterative difference" <|
                \() ->
                    Expect.equal (List.Extra.scanr1 (-) [ 1, 2, 3 ]) [ 2, -1, 3 ]
            , test "computes right to left flipped iterative difference" <|
                \() ->
                    Expect.equal (List.Extra.scanr1 (\x y -> y - x) [ 1, 2, 3 ]) [ 0, 1, 3 ]
            ]
        , describe "mapAccuml" <|
            [ test "on empty list" <|
                \() ->
                    Expect.equal
                        (List.Extra.mapAccuml (\a x -> ( a + x, a * x )) 5 [])
                        ( 5, [] )
            , test "accumulate sum and map product" <|
                \() ->
                    Expect.equal
                        (List.Extra.mapAccuml (\a x -> ( a + x, a * x )) 5 [ 2, 4, 8 ])
                        ( 19, [ 10, 28, 88 ] )
            , test "running total" <|
                \() ->
                    Expect.equal
                        (List.Extra.mapAccuml (\a x -> ( a + x, ( x, a + x ) )) 0 [ 2, 4, 8 ])
                        ( 14, [ ( 2, 2 ), ( 4, 6 ), ( 8, 14 ) ] )
            , test "works for very long list (i.e. is call stack size safe)" <|
                \() ->
                    Expect.equal
                        (List.Extra.mapAccuml (\a x -> ( a + x, () )) 0 (List.range 1 100000) |> Tuple.first)
                        5000050000
            ]
        , describe "mapAccumr" <|
            [ test "on empty list" <|
                \() ->
                    Expect.equal
                        (List.Extra.mapAccumr (\a x -> ( a + x, a * x )) 5 [])
                        ( 5, [] )
            , test "accumulate sum and map product" <|
                \() ->
                    Expect.equal
                        (List.Extra.mapAccumr (\a x -> ( a + x, a * x )) 5 [ 2, 4, 8 ])
                        ( 19, [ 34, 52, 40 ] )
            , test "add count down" <|
                \() ->
                    Expect.equal
                        (List.Extra.mapAccumr (\a x -> ( a + 1, ( x, a ) )) 0 [ 2, 4, 8 ])
                        ( 3, [ ( 2, 2 ), ( 4, 1 ), ( 8, 0 ) ] )
            , test "works for very long list (i.e. is call stack size safe)" <|
                \() ->
                    Expect.equal
                        (List.Extra.mapAccumr (\a x -> ( a + x, () )) 0 (List.range 1 100000) |> Tuple.first)
                        5000050000
            ]
        , describe "unfoldr" <|
            [ test "builds a decreasing list from a seed" <|
                \() ->
                    Expect.equal
                        (List.Extra.unfoldr
                            (\b ->
                                if b == 0 then
                                    Nothing

                                else
                                    Just ( b, b - 1 )
                            )
                            5
                        )
                        [ 5, 4, 3, 2, 1 ]
            ]
        , describe "iterate" <|
            [ test "collatz 13" <|
                \() ->
                    let
                        collatz n =
                            if n == 1 then
                                Nothing

                            else
                                Just <|
                                    if modBy 2 n == 0 then
                                        n // 2

                                    else
                                        3 * n + 1
                    in
                    Expect.equal (List.Extra.iterate collatz 13) [ 13, 40, 20, 10, 5, 16, 8, 4, 2, 1 ]
            , test "should not raise RangeError" <|
                \() ->
                    let
                        loop n =
                            if n == 100000 then
                                Nothing

                            else
                                Just (n + 1)

                        val =
                            List.Extra.iterate loop 1
                    in
                    Expect.equal (List.length val) 100000
            ]
        , describe "init" <|
            [ test "handles an empty list" <|
                \() ->
                    Expect.equal (List.Extra.init []) Nothing
            , fuzz Fuzz.int "handles a nearly-empty list" <|
                \x ->
                    Expect.equal (List.Extra.init [ x ]) (Just [])
            , fuzz2 (Fuzz.list Fuzz.int) Fuzz.int "handles a non-empty list" <|
                \list x ->
                    Expect.equal (List.Extra.init <| list ++ [ x ]) (Just list)
            ]
        , describe "initialize" <|
            [ test "creates a list starting from zero" <|
                \() ->
                    Expect.equal (List.Extra.initialize 5 identity) [ 0, 1, 2, 3, 4 ]
            , test "creates a list by doubling the index" <|
                \() ->
                    Expect.equal (List.Extra.initialize 5 (\x -> x * 2)) [ 0, 2, 4, 6, 8 ]
            , test "creates a list of identical values" <|
                \() ->
                    Expect.equal (List.Extra.initialize 1 (always 3)) [ 3 ]
            ]
        , describe "cycle" <|
            [ test "same length" <|
                \() ->
                    Expect.equal (List.Extra.cycle 3 [ 4, 7, 8 ]) [ 4, 7, 8 ]
            , test "multiple of length" <|
                \() ->
                    Expect.equal (List.Extra.cycle 6 [ 4, 7, 8 ]) [ 4, 7, 8, 4, 7, 8 ]
            , test "partial cycle" <|
                \() ->
                    Expect.equal (List.Extra.cycle 4 [ 'a', 'b', 'c' ]) [ 'a', 'b', 'c', 'a' ]
            , test "empty list" <|
                \() ->
                    Expect.equal (List.Extra.cycle 9001 []) []
            , test "resulting length smaller than cycle length" <|
                \() ->
                    Expect.equal (List.Extra.cycle 2 [ 1, 2, 3, 4, 5 ]) [ 1, 2 ]
            ]
        , describe "reverseRange"
            [ let
                rangeCeiling =
                    10000

                rangeFloor =
                    negate rangeCeiling
              in
              fuzz2 (Fuzz.intRange rangeFloor rangeCeiling) (Fuzz.intRange rangeFloor rangeCeiling) "always equal to the reverse of List.range" <|
                \hi lo ->
                    List.Extra.reverseRange hi lo
                        |> Expect.equalLists (List.reverse (List.range lo hi))
            ]
        , describe "splitAt" <|
            [ test "splits a list in the middle" <|
                \() ->
                    Expect.equal (List.Extra.splitAt 3 [ 1, 2, 3, 4, 5 ]) ( [ 1, 2, 3 ], [ 4, 5 ] )
            , test "splits a list at the first element" <|
                \() ->
                    Expect.equal (List.Extra.splitAt 1 [ 1, 2, 3 ]) ( [ 1 ], [ 2, 3 ] )
            , test "splits the entire list correctly" <|
                \() ->
                    Expect.equal (List.Extra.splitAt 3 [ 1, 2, 3 ]) ( [ 1, 2, 3 ], [] )
            , test "splits past the length of the list" <|
                \() ->
                    Expect.equal (List.Extra.splitAt 4 [ 1, 2, 3 ]) ( [ 1, 2, 3 ], [] )
            , test "handles zero correctly" <|
                \() ->
                    Expect.equal (List.Extra.splitAt 0 [ 1, 2, 3 ]) ( [], [ 1, 2, 3 ] )
            , test "handles negative numbers correctly" <|
                \() ->
                    Expect.equal (List.Extra.splitAt -1 [ 1, 2, 3 ]) ( [], [ 1, 2, 3 ] )
            ]
        , describe "splitWhen" <|
            [ test "returns split list when predicate is true" <|
                \() ->
                    Expect.equal (List.Extra.splitWhen (\n -> n == 3) [ 1, 2, 3, 4, 5 ]) (Just ( [ 1, 2 ], [ 3, 4, 5 ] ))
            , test "returns nothing when predicate is false" <|
                \() ->
                    Expect.equal (List.Extra.splitWhen (\n -> n == 6) [ 1, 2, 3, 4, 5 ]) Nothing
            ]
        , describe "takeWhileRight" <|
            [ test "keeps the correct items" <|
                \() ->
                    Expect.equal (List.Extra.takeWhileRight (\x -> 5 < x) (List.range 1 10)) [ 6, 7, 8, 9, 10 ]
            , test "drops the correct items" <|
                \() ->
                    Expect.equal (List.Extra.dropWhileRight (\x -> 5 < x) (List.range 1 10)) [ 1, 2, 3, 4, 5 ]
            ]
        , describe "takeWhile" <|
            [ test "doesn't exceed maximum call stack" <|
                \() ->
                    Expect.equal (List.Extra.takeWhile (\x -> 19999 > x) (List.range 1 20000)) (List.range 1 19998)
            ]
        , describe "span" <|
            [ test "splits in the middle of the list" <|
                \() ->
                    Expect.equal (List.Extra.span (\x -> 3 > x) [ 1, 2, 3, 4, 1, 2, 3, 4 ]) ( [ 1, 2 ], [ 3, 4, 1, 2, 3, 4 ] )
            , test "every element passes predicate" <|
                \() ->
                    Expect.equal (List.Extra.span (\x -> 5 > x) [ 1, 2, 3 ]) ( [ 1, 2, 3 ], [] )
            , test "first item doesn't pass predicate" <|
                \() ->
                    Expect.equal (List.Extra.span (\x -> 0 > x) [ 1, 2, 3 ]) ( [], [ 1, 2, 3 ] )
            ]
        , describe "break" <|
            [ test "breaks in the middle of the list" <|
                \() ->
                    Expect.equal (List.Extra.break (\x -> 3 < x) [ 1, 2, 3, 4, 1, 2, 3, 4 ]) ( [ 1, 2, 3 ], [ 4, 1, 2, 3, 4 ] )
            , test "breaks on the first item" <|
                \() ->
                    Expect.equal (List.Extra.break (\x -> 5 > x) [ 1, 2, 3 ]) ( [], [ 1, 2, 3 ] )
            , test "doesn't break for any element" <|
                \() ->
                    Expect.equal (List.Extra.break (\x -> 5 < x) [ 1, 2, 3 ]) ( [ 1, 2, 3 ], [] )
            ]
        , describe "stripPrefix" <|
            [ test "removes a matching prefix" <|
                \() ->
                    Expect.equal (List.Extra.stripPrefix [ 1, 2 ] [ 1, 2, 3, 4 ]) (Just [ 3, 4 ])
            , test "removes a matching 3-element prefix" <|
                \() ->
                    Expect.equal (List.Extra.stripPrefix [ 1, 2, 3 ] [ 1, 2, 3, 4, 5 ]) (Just [ 4, 5 ])
            , test "can remove the entire list" <|
                \() ->
                    Expect.equal (List.Extra.stripPrefix [ 1, 2, 3 ] [ 1, 2, 3 ]) (Just [])
            , test "fails when prefix is longer than list" <|
                \() ->
                    Expect.equal (List.Extra.stripPrefix [ 1, 2, 3 ] [ 1, 2 ]) Nothing
            , test "fails when list doesn't contain prefix" <|
                \() ->
                    Expect.equal (List.Extra.stripPrefix [ 3, 2, 1 ] [ 1, 2, 3, 4, 5 ]) Nothing
            ]
        , describe "group" <|
            [ test "groups elements correctly" <|
                \() ->
                    Expect.equal (List.Extra.group [ 1, 2, 2, 3, 3, 3, 2, 2, 1 ])
                        [ ( 1, [] ), ( 2, [ 2 ] ), ( 3, [ 3, 3 ] ), ( 2, [ 2 ] ), ( 1, [] ) ]
            ]
        , describe "groupWhile" <|
            [ test "groups by sub-element equality" <|
                \() ->
                    Expect.equal
                        (List.Extra.groupWhile (\( x, _ ) ( y, _ ) -> x == y) [ ( 0, 'a' ), ( 0, 'b' ), ( 1, 'c' ), ( 1, 'd' ) ])
                        [ ( ( 0, 'a' ), [ ( 0, 'b' ) ] ), ( ( 1, 'c' ), [ ( 1, 'd' ) ] ) ]
            , test "comparison function is reflexive, symmetric, and transitive" <|
                \() ->
                    Expect.equal
                        (List.Extra.groupWhile (<) [ 1, 2, 3, 2, 4, 1, 3, 2, 1 ])
                        [ ( 1, [ 2, 3 ] ), ( 2, [ 4 ] ), ( 1, [ 3 ] ), ( 2, [] ), ( 1, [] ) ]
            ]
        , describe "inits" <|
            [ test "returns all initial segments" <|
                \() ->
                    Expect.equal (List.Extra.inits [ 1, 2, 3 ]) [ [], [ 1 ], [ 1, 2 ], [ 1, 2, 3 ] ]
            ]
        , describe "tails" <|
            [ test "returns all final segments" <|
                \() ->
                    Expect.equal (List.Extra.tails [ 1, 2, 3 ]) [ [ 1, 2, 3 ], [ 2, 3 ], [ 3 ], [] ]
            ]
        , describe "select" <|
            [ test "returns all variations with a single item removed" <|
                \() ->
                    Expect.equal
                        (List.Extra.select [ 1, 2, 3, 4 ])
                        [ ( 1, [ 2, 3, 4 ] ), ( 2, [ 1, 3, 4 ] ), ( 3, [ 1, 2, 4 ] ), ( 4, [ 1, 2, 3 ] ) ]
            ]
        , describe "selectSplit" <|
            [ test "returns all splits at a single item" <|
                \() ->
                    Expect.equal (List.Extra.selectSplit [ 1, 2, 3 ]) [ ( [], 1, [ 2, 3 ] ), ( [ 1 ], 2, [ 3 ] ), ( [ 1, 2 ], 3, [] ) ]
            ]
        , describe "isSubsequenceOf" <|
            [ test "success" <|
                \() ->
                    List.Extra.isSubsequenceOf [ "E", "l", "m" ] [ "E", "a", "t", " ", "l", "i", "m", "e", "s" ]
                        |> Expect.equal True
                        |> Expect.onFail "Elm is a subsequence of Eat lime"
            , test "failure" <|
                \() ->
                    List.Extra.isSubsequenceOf [ "E", "l", "m" ] [ "E", "m", "a", "i", "l" ]
                        |> Expect.equal False
                        |> Expect.onFail "Elm is not a subsequence of Email"
            , test "success at last element" <|
                \() ->
                    List.Extra.isSubsequenceOf [ 1, 3 ] [ 1, 2, 3 ]
                        |> Expect.equal True
                        |> Expect.onFail "[] should be a subsequence of []"
            ]
        , describe "lift2" <|
            [ test "produces all combinations of addition" <|
                \() ->
                    Expect.equal (List.Extra.lift2 (+) [ 1, 2, 3 ] [ 4, 5 ]) [ 5, 6, 6, 7, 7, 8 ]
            ]
        , describe "groupsOf" <|
            [ test "groups by the correct number of items" <|
                \() ->
                    Expect.equal (List.Extra.groupsOf 3 (List.range 1 10))
                        [ [ 1, 2, 3 ], [ 4, 5, 6 ], [ 7, 8, 9 ] ]
            ]
        , describe "groupsOfWithStep" <|
            [ test "step == size" <|
                \() ->
                    Expect.equal (List.Extra.groupsOfWithStep 4 4 (List.range 1 10))
                        [ [ 1, 2, 3, 4 ], [ 5, 6, 7, 8 ] ]
            , test "step < size" <|
                \() ->
                    Expect.equal (List.Extra.groupsOfWithStep 3 1 (List.range 1 5))
                        [ [ 1, 2, 3 ], [ 2, 3, 4 ], [ 3, 4, 5 ] ]
            , test "step > size" <|
                \() ->
                    Expect.equal (List.Extra.groupsOfWithStep 3 6 (List.range 1 20))
                        [ [ 1, 2, 3 ], [ 7, 8, 9 ], [ 13, 14, 15 ] ]
            ]
        , describe "groupsOfVarying" <|
            [ test "group sizes match passed sizes" <|
                \() ->
                    Expect.equal
                        (List.Extra.groupsOfVarying [ 2, 3, 1 ] [ "a", "b", "c", "d", "e", "f" ])
                        [ [ "a", "b" ], [ "c", "d", "e" ], [ "f" ] ]
            , test "groups correctly when passed counts are less than list size" <|
                \() ->
                    Expect.equal
                        (List.Extra.groupsOfVarying [ 2 ] [ "a", "b", "c", "d", "e", "f" ])
                        [ [ "a", "b" ] ]
            , test "groups correctly when passed counts are greater than list size" <|
                \() ->
                    Expect.equal
                        (List.Extra.groupsOfVarying [ 2, 3, 1, 5, 6 ] [ "a", "b", "c", "d", "e" ])
                        [ [ "a", "b" ], [ "c", "d", "e" ] ]
            ]
        , describe "greedyGroupsOf" <|
            [ test "groups correctly while keeping trailing group" <|
                \() ->
                    Expect.equal (List.Extra.greedyGroupsOf 3 (List.range 1 10))
                        [ [ 1, 2, 3 ], [ 4, 5, 6 ], [ 7, 8, 9 ], [ 10 ] ]
            ]
        , describe "greedyGroupsOfWithStep" <|
            [ test "step == size" <|
                \() ->
                    Expect.equal (List.Extra.greedyGroupsOfWithStep 4 4 (List.range 1 10))
                        [ [ 1, 2, 3, 4 ], [ 5, 6, 7, 8 ], [ 9, 10 ] ]
            , test "step < size" <|
                \() ->
                    Expect.equal (List.Extra.greedyGroupsOfWithStep 3 2 (List.range 1 6))
                        [ [ 1, 2, 3 ], [ 3, 4, 5 ], [ 5, 6 ] ]
            , test "step > size" <|
                \() ->
                    Expect.equal (List.Extra.greedyGroupsOfWithStep 3 6 (List.range 1 20))
                        [ [ 1, 2, 3 ], [ 7, 8, 9 ], [ 13, 14, 15 ], [ 19, 20 ] ]
            ]
        , describe "isPrefixOf"
            [ fuzz (Fuzz.list Fuzz.int) "[] is prefix to anything" <|
                \list ->
                    List.Extra.isPrefixOf [] list
                        |> Expect.equal True
                        |> Expect.onFail "Expected [] to be a prefix."
            , fuzz (Fuzz.list Fuzz.int) "reflexivity" <|
                \list ->
                    List.Extra.isPrefixOf list list
                        |> Expect.equal True
                        |> Expect.onFail "Expected list to be a prefix of itself."
            , fuzz2 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "antisymmetry" <|
                \listA listB ->
                    not (List.Extra.isPrefixOf listA listB)
                        || not (List.Extra.isPrefixOf listB listA)
                        || listA
                        == listB
                        |> Expect.equal True
                        |> Expect.onFail "Expected exactly one to be prefix of the other."
            , fuzz3 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "transitivity" <|
                \listA listB listC ->
                    not (List.Extra.isPrefixOf listA listB)
                        || not (List.Extra.isPrefixOf listB listC)
                        || List.Extra.isPrefixOf listA listC
                        |> Expect.equal True
                        |> Expect.onFail "Expected prefix of prefix to be prefix."
            , test "stack safety" <|
                \() ->
                    List.Extra.isPrefixOf (List.range 1 6000) (List.range 1 10000)
                        |> Expect.equal True
                        |> Expect.onFail "1, 2, ..., 6k is prefix of 1, 2, ..., 10k"
            , fuzz2 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "generalized fuzz test" <|
                \prefix rest ->
                    List.Extra.isPrefixOf prefix (prefix ++ rest)
                        |> Expect.equal True
                        |> Expect.onFail "xs is prefix of xs ++ ys"
            ]
        , describe "isSuffixOf"
            [ fuzz (Fuzz.list Fuzz.int) "[] is suffix to anything" <|
                \list ->
                    List.Extra.isSuffixOf [] list
                        |> Expect.equal True
                        |> Expect.onFail "Expected [] to be a suffix."
            , fuzz (Fuzz.list Fuzz.int) "reflexivity" <|
                \list ->
                    List.Extra.isSuffixOf list list
                        |> Expect.equal True
                        |> Expect.onFail "Expected list to be a suffix of itself."
            , fuzz2 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "antisymmetry" <|
                \listA listB ->
                    not (List.Extra.isSuffixOf listA listB)
                        || not (List.Extra.isSuffixOf listB listA)
                        || listA
                        == listB
                        |> Expect.equal True
                        |> Expect.onFail "Expected exactly one to be suffix of the other."
            , fuzz3 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "transitivity" <|
                \listA listB listC ->
                    not (List.Extra.isSuffixOf listA listB)
                        || not (List.Extra.isSuffixOf listB listC)
                        || List.Extra.isSuffixOf listA listC
                        |> Expect.equal True
                        |> Expect.onFail "Expected suffix of suffix to be suffix."
            , test "stack safety" <|
                \() ->
                    List.Extra.isSuffixOf (List.range 4000 10000) (List.range 1 10000)
                        |> Expect.equal True
                        |> Expect.onFail "4000, 4001, ..., 10k is suffix of 1, 2, ..., 10k"
            , fuzz2 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "generalized fuzz test" <|
                \suffix rest ->
                    List.Extra.isSuffixOf suffix (rest ++ suffix)
                        |> Expect.equal True
                        |> Expect.onFail "ys is suffix of xs ++ ys"
            ]
        , describe "isInfixOf"
            [ test "success" <|
                \() ->
                    List.Extra.isInfixOf [ 5, 7, 11 ] [ 2, 3, 5, 7, 11, 13 ]
                        |> Expect.equal True
                        |> Expect.onFail "5, 7, 11 is infix of 2, 3, 5, 7, 11, 13"
            , test "not consecutive" <|
                \() ->
                    List.Extra.isInfixOf [ 5, 7, 13 ] [ 2, 3, 5, 7, 11, 13 ]
                        |> Expect.equal False
                        |> Expect.onFail "5, 7, 13 is not infix of 2, 3, 5, 7, 11, 13"
            , test "not in-order" <|
                \() ->
                    List.Extra.isInfixOf [ 3, 5, 2 ] [ 2, 3, 5, 7, 11, 13 ]
                        |> Expect.equal False
                        |> Expect.onFail "3, 5, 2 is not infix of 2, 3, 5, 7, 11, 13"
            , test "partial match then real match" <|
                \() ->
                    List.Extra.isInfixOf [ 1, 2 ] [ 1, 3, 1, 2 ]
                        |> Expect.equal True
                        |> Expect.onFail "1, 2 is infix of 1, 3, 1, 2"
            , test "stack safety" <|
                \() ->
                    List.Extra.isInfixOf (List.range 1 6000) (List.range 1 10000)
                        |> Expect.equal True
                        |> Expect.onFail "1, 2, ..., 6k is infix of 1, 2, ..., 10k"
            , fuzz3 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "generalized fuzz test" <|
                \prefix match suffix ->
                    List.Extra.isInfixOf match (prefix ++ match ++ suffix)
                        |> Expect.equal True
                        |> Expect.onFail "ys is infix of xs ++ ys ++ zs"
            , test "empty is infix of empty" <|
                \() ->
                    List.Extra.isInfixOf [] []
                        |> Expect.equal True
                        |> Expect.onFail "empty is infix of empty"
            , fuzz (Fuzz.list Fuzz.int) "empty is infix of anything" <|
                \list ->
                    List.Extra.isInfixOf [] list
                        |> Expect.equal True
                        |> Expect.onFail "empty is infix of anything"
            , fuzz2 Fuzz.int (Fuzz.list Fuzz.int) "non-empty is not infix of empty" <|
                \x xs ->
                    List.Extra.isInfixOf (x :: xs) []
                        |> Expect.equal False
                        |> Expect.onFail "non-empty is not infix of empty"
            , fuzz (Fuzz.list Fuzz.int) "equal lists are infix" <|
                \list ->
                    List.Extra.isInfixOf list list
                        |> Expect.equal True
                        |> Expect.onFail "equal lists are infix"
            , test "is stack safe" <|
                \() ->
                    List.Extra.isInfixOf [ 5, 7, 13 ] (List.repeat 1000000 5)
                        |> Expect.equal False
                        |> Expect.onFail "5, 7, 13 is not infix of 2, 3, 5, 7, 11, 13"
            ]
        , describe "swapAt"
            [ test "negative index as first argument returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.swapAt -1 0 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            , test "negative index as second argument returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.swapAt 0 -1 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            , test "out of range index as first argument returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.swapAt 10 0 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            , test "out of range index as second argument returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.swapAt 0 -1 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            , test "identical indexes returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.swapAt 1 1 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            , test "swap the elements at indices 0 and 1" <|
                \() ->
                    Expect.equal (List.Extra.swapAt 0 1 [ 1, 2, 3 ]) [ 2, 1, 3 ]
            ]
        , describe "removeAt"
            [ test "negative index returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.removeAt -1 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            , test "remove the element at index 0" <|
                \() ->
                    Expect.equal (List.Extra.removeAt 0 [ 1, 2, 3 ]) [ 2, 3 ]
            , test "remove the element at index 2" <|
                \() ->
                    Expect.equal (List.Extra.removeAt 2 [ 1, 2, 3 ]) [ 1, 2 ]
            , test "out of range index returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.removeAt 4 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            ]
        , describe "setAt"
            [ test "negative index returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.setAt -1 9 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            , test "set index 0 to 9" <|
                \() ->
                    Expect.equal (List.Extra.setAt 0 9 [ 1, 2, 3 ]) [ 9, 2, 3 ]
            , test "set index 2 to 9" <|
                \() ->
                    Expect.equal (List.Extra.setAt 2 9 [ 1, 2, 3 ]) [ 1, 2, 9 ]
            , test "out of range index returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.setAt 4 9 [ 1, 2, 3 ]) [ 1, 2, 3 ]
            ]
        , describe "updateAt"
            [ test "negative index returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.updateAt -1 ((+) 1) [ 1, 2, 3 ]) [ 1, 2, 3 ]
            , test "increment the element at index 0" <|
                \() ->
                    Expect.equal (List.Extra.updateAt 0 ((+) 1) [ 1, 2, 3 ]) [ 2, 2, 3 ]
            , test "increment the element at index 2" <|
                \() ->
                    Expect.equal (List.Extra.updateAt 2 ((+) 1) [ 1, 2, 3 ]) [ 1, 2, 4 ]
            , test "out of range index returns the original list" <|
                \() ->
                    Expect.equal (List.Extra.updateAt 4 ((+) 1) [ 1, 2, 3 ]) [ 1, 2, 3 ]
            ]
        , describe "updateIfIndex"
            [ test "increments first element" <|
                \() ->
                    Expect.equal (List.Extra.updateIfIndex (always True) ((+) 1) [ 1, 2, 3 ]) [ 2, 3, 4 ]
            , test "if the index is 2, then increment the element" <|
                \() ->
                    Expect.equal (List.Extra.updateIfIndex ((==) 2) ((+) 1) [ 1, 2, 3 ]) [ 1, 2, 4 ]
            , test "if the index is even, increment the element" <|
                \() ->
                    Expect.equal (List.Extra.updateIfIndex (\index -> modBy 2 index == 0) ((+) 1) [ 1, 2, 3 ]) [ 2, 2, 4 ]
            ]
        , describe "remove"
            [ test "leaves the list untouched if the value is not in the list" <|
                \() ->
                    let
                        list : List Int
                        list =
                            [ 1, 2, 3 ]
                    in
                    list
                        |> List.Extra.remove 1000
                        |> Expect.equal list
            , test "removes the element if it's present in the list" <|
                \() ->
                    [ 1, 2, 3 ]
                        |> List.Extra.remove 2
                        |> Expect.equal [ 1, 3 ]
            , test "removes only the first element" <|
                \() ->
                    [ 1, 2, 3, 3, 3 ]
                        |> List.Extra.remove 3
                        |> Expect.equal [ 1, 2, 3, 3 ]
            , test "is stack-safe" <|
                \() ->
                    let
                        list : List Int
                        list =
                            List.range 1 10000
                    in
                    list
                        |> List.Extra.remove 100000
                        |> Expect.equal list
            ]
        , describe "removeIfIndex"
            [ test "remove all the elements" <|
                \() ->
                    Expect.equal (List.Extra.removeIfIndex (always True) [ 1, 2, 3 ]) []
            , test "remove the element at index 2" <|
                \() ->
                    Expect.equal (List.Extra.removeIfIndex ((==) 2) [ 1, 2, 3 ]) [ 1, 2 ]
            , test "remove all elements at even indices" <|
                \() ->
                    Expect.equal (List.Extra.removeIfIndex (\index -> modBy 2 index == 0) [ 1, 2, 3 ]) [ 2 ]
            ]
        , describe "unconsLast"
            [ test "removes last element of list" <|
                \() ->
                    Expect.equal (List.Extra.unconsLast [ 1, 2, 3 ]) (Just ( 3, [ 1, 2 ] ))
            , test "returns Nothing if the list is empty" <|
                \() ->
                    Expect.equal (List.Extra.unconsLast []) Nothing
            ]
        , describe "maximumWith"
            [ test "maximum of empty list" <|
                \() ->
                    Expect.equal (List.Extra.maximumWith compare []) Nothing
            , test "first maximum of records list" <|
                \() ->
                    Expect.equal
                        (List.Extra.maximumWith (\x y -> compare x.val y.val) [ { id = 1, val = 1 }, { id = 2, val = 2 }, { id = 3, val = 2 } ])
                        (Just { id = 2, val = 2 })
            ]
        , describe "maximumBy"
            [ test "maximumBy of empty list" <|
                \() -> Expect.equal (List.Extra.maximumBy (\x -> x) []) Nothing
            , test "first maximumBy of records list" <|
                \() ->
                    Expect.equal (List.Extra.maximumBy (\x -> x.val) [ { id = 1, val = 1 }, { id = 2, val = 2 }, { id = 3, val = 2 } ])
                        (Just { id = 2, val = 2 })
            ]
        , describe "minimumWith"
            [ test "minimum of empty list" <|
                \() ->
                    Expect.equal (List.Extra.minimumWith compare []) Nothing
            , test "first minimum of records list" <|
                \() ->
                    Expect.equal
                        (List.Extra.minimumWith (\x y -> compare x.val y.val) [ { id = 1, val = 2 }, { id = 2, val = 1 }, { id = 3, val = 1 } ])
                        (Just { id = 2, val = 1 })
            ]
        , describe "minimumBy"
            [ test "minimumBy of empty list" <|
                \() -> Expect.equal (List.Extra.minimumBy (\x -> x) []) Nothing
            , test "first minimumBy of records list" <|
                \() ->
                    Expect.equal (List.Extra.minimumBy (\x -> x.val) [ { id = 1, val = 2 }, { id = 2, val = 1 }, { id = 3, val = 1 } ])
                        (Just { id = 2, val = 1 })
            ]
        , describe "setIf"
            [ test "empty list" <|
                \() ->
                    Expect.equal (List.Extra.setIf ((==) 1) 0 []) []
            , test "set all" <|
                \() ->
                    Expect.equal (List.Extra.setIf (always True) 2 [ 1, 2, 3, 4 ]) [ 2, 2, 2, 2 ]
            , test "set only evens" <|
                \() ->
                    Expect.equal (List.Extra.setIf (\x -> modBy 2 x == 0) 0 [ 17, 8, 2, 9 ]) [ 17, 0, 0, 9 ]
            ]
        , describe "gatherEquals"
            [ test "empty list" <|
                \() ->
                    List.Extra.gatherEquals []
                        |> Expect.equal []
            , test "single element" <|
                \() ->
                    List.Extra.gatherEquals [ 1 ]
                        |> Expect.equal [ ( 1, [] ) ]
            , test "proper test" <|
                \() ->
                    List.Extra.gatherEquals [ 1, 2, 1, 2, 3, 4, 1 ]
                        |> Expect.equal
                            [ ( 1, [ 1, 1 ] )
                            , ( 2, [ 2 ] )
                            , ( 3, [] )
                            , ( 4, [] )
                            ]
            ]
        , describe "gatherEqualsBy"
            [ test "empty list" <|
                \() ->
                    List.Extra.gatherEqualsBy identity []
                        |> Expect.equal []
            , test "single element" <|
                \() ->
                    List.Extra.gatherEqualsBy identity [ 1 ]
                        |> Expect.equal [ ( 1, [] ) ]
            , test "proper test" <|
                \() ->
                    List.Extra.gatherEqualsBy identity [ 1, 2, 1, 2, 3, 4, 1 ]
                        |> Expect.equal
                            [ ( 1, [ 1, 1 ] )
                            , ( 2, [ 2 ] )
                            , ( 3, [] )
                            , ( 4, [] )
                            ]
            ]
        , describe "gatherWith"
            [ test "empty list" <|
                \() ->
                    List.Extra.gatherWith (==) []
                        |> Expect.equal []
            , test "single element" <|
                \() ->
                    List.Extra.gatherWith (==) [ 1 ]
                        |> Expect.equal [ ( 1, [] ) ]
            , test "proper test" <|
                \() ->
                    List.Extra.gatherWith (==) [ 1, 2, 1, 2, 3, 4, 1 ]
                        |> Expect.equal
                            [ ( 1, [ 1, 1 ] )
                            , ( 2, [ 2 ] )
                            , ( 3, [] )
                            , ( 4, [] )
                            ]
            ]
        , describe "uniquePairs"
            [ test "empty list" <|
                \() ->
                    List.Extra.uniquePairs []
                        |> Expect.equal []
            , test "single element has no counterpart to pair with" <|
                \() ->
                    List.Extra.uniquePairs [ 1 ]
                        |> Expect.equal []
            , test "two elements have exactly one way to pair" <|
                \() ->
                    List.Extra.uniquePairs [ 1, 2 ]
                        |> Expect.equal [ ( 1, 2 ) ]
            , test "three elements have three ways to pair" <|
                \() ->
                    List.Extra.uniquePairs [ 1, 2, 3 ]
                        |> Expect.equal [ ( 1, 2 ), ( 1, 3 ), ( 2, 3 ) ]
            ]
        , describe "joinOn"
            [ test "with first list empty" <|
                \() ->
                    List.Extra.joinOn Tuple.pair identity identity [] [ 1, 2, 3 ]
                        |> Expect.equal []
            , test "with second list empty" <|
                \() ->
                    List.Extra.joinOn Tuple.pair identity identity [ 1, 2, 3 ] []
                        |> Expect.equal []
            , test "with neither list empty" <|
                \() ->
                    List.Extra.joinOn Tuple.pair identity identity [ 1, 3, 2 ] [ 2, 1, 3 ]
                        |> Expect.equal [ ( 3, 3 ), ( 2, 2 ), ( 1, 1 ) ]
            ]
        , describe "frequencies"
            [ describe "Property testing with fuzz"
                [ fuzz (Fuzz.list Fuzz.int) "Final value similar list and List.sort list" <|
                    \list ->
                        List.Extra.frequencies list
                            |> List.sort
                            |> Expect.equal (List.Extra.frequencies list)
                , fuzz (Fuzz.list Fuzz.int) "Final value similar, no matter the order of composition for frequencies and List.sort" <|
                    \list ->
                        let
                            freqFirst =
                                list
                                    |> List.Extra.frequencies
                                    |> List.sort

                            sortFirst =
                                list
                                    |> List.sort
                                    |> List.Extra.frequencies
                        in
                        freqFirst
                            |> Expect.equal sortFirst
                , fuzz2 (Fuzz.list Fuzz.int) (Fuzz.list Fuzz.int) "Two (different/similar) lists should give 2 (different/similar) results" <|
                    \list1 list2 ->
                        let
                            areListsSimilar =
                                List.Extra.isPermutationOf list2 list1

                            freq1 =
                                List.Extra.frequencies list1

                            freq2 =
                                List.Extra.frequencies list2

                            shouldFreqsBeSimilar =
                                areListsSimilar

                            areFreqsSimilar =
                                List.Extra.isPermutationOf freq2 freq1
                        in
                        areFreqsSimilar
                            |> Expect.equal shouldFreqsBeSimilar
                ]
            , describe "Unit testing on basic examples"
                [ test "Frequencies on List Int" <|
                    \() ->
                        List.Extra.frequencies [ 4, 1, 3, 2, 2, 4, 3, 3, 4, 4 ]
                            |> Expect.equal [ ( 1, 1 ), ( 2, 2 ), ( 3, 3 ), ( 4, 4 ) ]
                , test "Frequencies on List String" <|
                    \() ->
                        List.Extra.frequencies [ "a", "b", "aa", "c", "b", "aa", "c", "C", "C", "D" ]
                            |> Expect.equal [ ( "C", 2 ), ( "D", 1 ), ( "a", 1 ), ( "aa", 2 ), ( "b", 2 ), ( "c", 2 ) ]
                , test "Frequencies on empty List a" <|
                    \() ->
                        List.Extra.frequencies []
                            |> Expect.equal []
                ]
            ]
        , describe "stoppableFoldl"
            [ fuzz (Fuzz.list Fuzz.int) "behaves like foldl if function always returns Continue" <|
                \xs ->
                    List.Extra.stoppableFoldl (\n acc -> Continue (n + acc)) 0 xs
                        |> Expect.equal (List.sum xs)
            , test "simple example" <|
                \() ->
                    List.Extra.stoppableFoldl
                        (\n acc ->
                            if acc >= 50 then
                                Stop acc

                            else
                                Continue (n + acc)
                        )
                        0
                        (List.range 1 1000)
                        |> Expect.equal 55
            , test "after Stop, the function is not called anymore" <|
                let
                    throwRangeErrorException : () -> a
                    throwRangeErrorException () =
                        preventTailCallOptimization ()

                    preventTailCallOptimization : () -> a
                    preventTailCallOptimization () =
                        throwRangeErrorException ()
                in
                \() ->
                    List.Extra.stoppableFoldl
                        (\n _ ->
                            if n < 50 then
                                Continue n

                            else if n == 50 then
                                Stop n

                            else
                                throwRangeErrorException ()
                        )
                        0
                        (List.range 1 1000)
                        |> Expect.equal 50
            ]
        ]
